@@ -41,6 +41,8 @@ G_BEGIN_DECLS
 
 #define GST_H264_MAX_SPS_COUNT   32
 #define GST_H264_MAX_PPS_COUNT   256
+#define GST_H264_MAX_VIEW_COUNT  1024
+#define GST_H264_MAX_VIEW_ID     (GST_H264_MAX_VIEW_COUNT - 1)
 
 #define GST_H264_IS_P_SLICE(slice)  (((slice)->type % 5) == GST_H264_P_SLICE)
 #define GST_H264_IS_B_SLICE(slice)  (((slice)->type % 5) == GST_H264_B_SLICE)
@@ -48,6 +50,11 @@ G_BEGIN_DECLS
 #define GST_H264_IS_SP_SLICE(slice) (((slice)->type % 5) == GST_H264_SP_SLICE)
 #define GST_H264_IS_SI_SLICE(slice) (((slice)->type % 5) == GST_H264_SI_SLICE)
 
+#define GST_H264_IS_SVC_NALU(nalu) \
+  ((nalu)->extension_type == GST_H264_NAL_EXTENSION_SVC)
+#define GST_H264_IS_MVC_NALU(nalu) \
+  ((nalu)->extension_type == GST_H264_NAL_EXTENSION_MVC)
+
 /**
  * GstH264Profile:
  * @GST_H264_PROFILE_BASELINE: Baseline profile (A.2.1)
@@ -101,6 +108,11 @@ typedef enum {
  * @GST_H264_NAL_SEQ_END: End of sequence nal unit
  * @GST_H264_NAL_STREAM_END: End of stream nal unit
  * @GST_H264_NAL_FILLER_DATA: Filler data nal lunit
+ * @GST_H264_NAL_SPS_EXT: Sequence parameter set (SPS) extension NAL unit
+ * @GST_H264_NAL_PREFIX_UNIT: Prefix NAL unit
+ * @GST_H264_NAL_SUBSET_SPS: Subset sequence parameter set (SPS) NAL unit
+ * @GST_H264_NAL_SLICE_AUX: Auxiliary coded picture without partitioning NAL unit
+ * @GST_H264_NAL_SLICE_EXT: Coded slice extension NAL unit
  *
  * Indicates the type of H264 Nal Units
  */
@@ -118,9 +130,29 @@ typedef enum
   GST_H264_NAL_AU_DELIMITER = 9,
   GST_H264_NAL_SEQ_END      = 10,
   GST_H264_NAL_STREAM_END   = 11,
-  GST_H264_NAL_FILLER_DATA  = 12
+  GST_H264_NAL_FILLER_DATA  = 12,
+  GST_H264_NAL_SPS_EXT      = 13,
+  GST_H264_NAL_PREFIX_UNIT  = 14,
+  GST_H264_NAL_SUBSET_SPS   = 15,
+  GST_H264_NAL_SLICE_AUX    = 19,
+  GST_H264_NAL_SLICE_EXT    = 20
 } GstH264NalUnitType;
 
+/**
+ * GstH264NalUnitExtensionType:
+ * @GST_H264_NAL_EXTENSION_NONE: No NAL unit header extension is available
+ * @GST_H264_NAL_EXTENSION_SVC: NAL unit header extension for SVC (Annex G)
+ * @GST_H264_NAL_EXTENSION_MVC: NAL unit header extension for MVC (Annex H)
+ *
+ * Indicates the type of H.264 NAL unit extension.
+ */
+typedef enum
+{
+  GST_H264_NAL_EXTENSION_NONE = 0,
+  GST_H264_NAL_EXTENSION_SVC,
+  GST_H264_NAL_EXTENSION_MVC,
+} GstH264NalUnitExtensionType;
+
 /**
  * GstH264ParserResult:
  * @GST_H264_PARSER_OK: The parsing succeded
@@ -142,10 +174,37 @@ typedef enum
   GST_H264_PARSER_NO_NAL_END
 } GstH264ParserResult;
 
+/**
+ * GstH264FramePackingType:
+ * @GST_H264_FRAME_PACKING_NONE: A complete 2D frame without any frame packing
+ * @GST_H264_FRAME_PACKING_CHECKERBOARD_INTERLEAVING: Checkerboard
+ *   based interleaving
+ * @GST_H264_FRAME_PACKING_COLUMN_INTERLEAVING: Column based interleaving
+ * @GST_H264_FRAME_PACKING_ROW_INTERLEAVING: Row based interleaving
+ * @GST_H264_FRAME_PACKING_SIDE_BY_SIDE: Side-by-side packing
+ * @GST_H264_FRMAE_PACKING_TOP_BOTTOM: Top-Bottom packing
+ * @GST_H264_FRAME_PACKING_TEMPORAL_INTERLEAVING: Temporal interleaving
+ *
+ * Frame packing arrangement types.
+ */
+typedef enum
+{
+  GST_H264_FRAME_PACKING_NONE                           = 6,
+  GST_H264_FRAME_PACKING_CHECKERBOARD_INTERLEAVING      = 0,
+  GST_H264_FRAME_PACKING_COLUMN_INTERLEAVING            = 1,
+  GST_H264_FRAME_PACKING_ROW_INTERLEAVING               = 2,
+  GST_H264_FRAME_PACKING_SIDE_BY_SIDE                   = 3,
+  GST_H264_FRMAE_PACKING_TOP_BOTTOM                     = 4,
+  GST_H264_FRAME_PACKING_TEMPORAL_INTERLEAVING          = 5
+} GstH264FramePackingType;
+
 /**
  * GstH264SEIPayloadType:
  * @GST_H264_SEI_BUF_PERIOD: Buffering Period SEI Message
  * @GST_H264_SEI_PIC_TIMING: Picture Timing SEI Message
+ * @GST_H264_SEI_RECOVERY_POINT: Recovery Point SEI Message (D.2.7)
+ * @GST_H264_SEI_STEREO_VIDEO_INFO: Stereo Video Information SEI Message (D.2.22)
+ * @GST_H264_SEI_FRAME_PACKING: Frame Packing Arrangement SEI Message (D.2.25)
  * ...
  *
  * The type of SEI message.
@@ -153,7 +212,10 @@ typedef enum
 typedef enum
 {
   GST_H264_SEI_BUF_PERIOD = 0,
-  GST_H264_SEI_PIC_TIMING = 1
+  GST_H264_SEI_PIC_TIMING = 1,
+  GST_H264_SEI_RECOVERY_POINT = 6,
+  GST_H264_SEI_STEREO_VIDEO_INFO = 21,
+  GST_H264_SEI_FRAME_PACKING = 45
       /* and more...  */
 } GstH264SEIPayloadType;
 
@@ -209,6 +271,12 @@ typedef enum
 typedef struct _GstH264NalParser              GstH264NalParser;
 
 typedef struct _GstH264NalUnit                GstH264NalUnit;
+typedef struct _GstH264NalUnitExtensionMVC    GstH264NalUnitExtensionMVC;
+
+typedef struct _GstH264SPSExtMVCView          GstH264SPSExtMVCView;
+typedef struct _GstH264SPSExtMVCLevelValue    GstH264SPSExtMVCLevelValue;
+typedef struct _GstH264SPSExtMVCLevelValueOp  GstH264SPSExtMVCLevelValueOp;
+typedef struct _GstH264SPSExtMVC              GstH264SPSExtMVC;
 
 typedef struct _GstH264SPS                    GstH264SPS;
 typedef struct _GstH264PPS                    GstH264PPS;
@@ -224,22 +292,53 @@ typedef struct _GstH264SliceHdr               GstH264SliceHdr;
 typedef struct _GstH264ClockTimestamp         GstH264ClockTimestamp;
 typedef struct _GstH264PicTiming              GstH264PicTiming;
 typedef struct _GstH264BufferingPeriod        GstH264BufferingPeriod;
+typedef struct _GstH264RecoveryPoint          GstH264RecoveryPoint;
+typedef struct _GstH264StereoVideoInfo        GstH264StereoVideoInfo;
+typedef struct _GstH264FramePacking           GstH264FramePacking;
 typedef struct _GstH264SEIMessage             GstH264SEIMessage;
 
+/**
+ * GstH264NalUnitExtensionMVC:
+ * @non_idr_flag: If equal to 0, it specifies that the current access
+ *   unit is an IDR access unit
+ * @priority_id: The priority identifier for the NAL unit
+ * @view_id: The view identifier for the NAL unit
+ * @temporal_id: The temporal identifier for the NAL unit
+ * @anchor_pic_flag: If equal to 1, it specifies that the current
+ *   access unit is an anchor access unit
+ * @inter_view_flag: If equal to 0, it specifies that the current view
+ *   component is not used for inter-view prediction by any other view
+ *   component in the current access unit
+ */
+struct _GstH264NalUnitExtensionMVC
+{
+  guint8 non_idr_flag;
+  guint8 priority_id;
+  guint16 view_id;
+  guint8 temporal_id;
+  guint8 anchor_pic_flag;
+  guint8 inter_view_flag;
+};
+
 /**
  * GstH264NalUnit:
- * @ref_idc: not equal to 0 specifies that the content of the NAL unit contains a sequence
- *  parameter set, a sequence * parameter set extension, a subset sequence parameter set, a
- *  picture parameter set, a slice of a reference picture, a slice data partition of a
- *  reference picture, or a prefix NAL unit preceding a slice of a reference picture.
+ * @ref_idc: not equal to 0 specifies that the content of the NAL unit
+ *  contains a sequence parameter set, a sequence parameter set
+ *  extension, a subset sequence parameter set, a picture parameter
+ *  set, a slice of a reference picture, a slice data partition of a
+ *  reference picture, or a prefix NAL unit preceding a slice of a
+ *  reference picture.
  * @type: A #GstH264NalUnitType
  * @idr_pic_flag: calculated idr_pic_flag
- * @size: The size of the nal unit starting from @offset
- * @offset: The offset of the actual start of the nal unit
- * @sc_offset:The offset of the start code of the nal unit
- * @valid: If the nal unit is valid, which mean it has
+ * @size: The size of the nal unit starting from @offset, thus
+ *  including the header bytes. e.g. @type (nal_unit_type)
+ * @offset: The offset of the actual start of the nal unit, thus
+ *  including the header bytes
+ * @sc_offset: The offset of the start code of the nal unit
+ * @valid: If the nal unit is valid, which means it has
  * already been parsed
  * @data: The data from which the Nalu has been parsed
+ * @header_bytes: The size of the NALU header in bytes
  *
  * Structure defining the Nal unit headers
  */
@@ -256,6 +355,12 @@ struct _GstH264NalUnit
   gboolean valid;
 
   guint8 *data;
+
+  guint8 header_bytes;
+  guint8 extension_type;
+  union {
+    GstH264NalUnitExtensionMVC mvc;
+  } extension;
 };
 
 /**
@@ -385,7 +490,7 @@ struct _GstH264VUIParams
   GstH264HRDParams nal_hrd_parameters;
 
   guint8 vcl_hrd_parameters_present_flag;
-  /* if nal_hrd_parameters_present_flag */
+  /* if vcl_hrd_parameters_present_flag */
   GstH264HRDParams vcl_hrd_parameters;
 
   guint8 low_delay_hrd_flag;
@@ -406,6 +511,97 @@ struct _GstH264VUIParams
   guint par_d;
 };
 
+/**
+ * GstH264SPSExtMVCView:
+ * @num_anchor_refs_l0: specifies the number of view components for
+ *   inter-view prediction in the initialized RefPicList0 in decoding
+ *   anchor view components.
+ * @anchor_ref_l0: specifies the view_id for inter-view prediction in
+ *   the initialized RefPicList0 in decoding anchor view components.
+ * @num_anchor_refs_l1: specifies the number of view components for
+ *   inter-view prediction in the initialized RefPicList1 in decoding
+ *   anchor view components.
+ * @anchor_ref_l1: specifies the view_id for inter-view prediction in
+ *   the initialized RefPicList1 in decoding anchor view components.
+ * @num_non_anchor_refs_l0: specifies the number of view components
+ *   for inter-view prediction in the initialized RefPicList0 in
+ *   decoding non-anchor view components.
+ * @non_anchor_ref_l0: specifies the view_id for inter-view prediction
+ *   in the initialized RefPicList0 in decoding non-anchor view
+ *   components.
+ * @num_non_anchor_refs_l1: specifies the number of view components
+ *   for inter-view prediction in the initialized RefPicList1 in
+ *   decoding non-anchor view components.
+ * @non_anchor_ref_l1: specifies the view_id for inter-view prediction
+ *   in the initialized RefPicList1 in decoding non-anchor view
+ *   components.
+ *
+ * Represents inter-view dependency relationships for the coded video
+ * sequence.
+ */
+struct _GstH264SPSExtMVCView
+{
+  guint16 view_id;
+  guint8 num_anchor_refs_l0;
+  guint16 anchor_ref_l0[15];
+  guint8 num_anchor_refs_l1;
+  guint16 anchor_ref_l1[15];
+  guint8 num_non_anchor_refs_l0;
+  guint16 non_anchor_ref_l0[15];
+  guint8 num_non_anchor_refs_l1;
+  guint16 non_anchor_ref_l1[15];
+};
+
+/**
+ * GstH264SPSExtMVCLevelValueOp:
+ *
+ * Represents an operation point for the coded video sequence.
+ */
+struct _GstH264SPSExtMVCLevelValueOp
+{
+  guint8 temporal_id;
+  guint16 num_target_views_minus1;
+  guint16 *target_view_id;
+  guint16 num_views_minus1;
+};
+
+/**
+ * GstH264SPSExtMVCLevelValue:
+ * @level_idc: specifies the level value signalled for the coded video
+ *   sequence
+ * @num_applicable_ops_minus1: plus 1 specifies the number of
+ *   operation points to which the level indicated by level_idc applies
+ * @applicable_op: specifies the applicable operation point
+ *
+ * Represents level values for a subset of the operation points for
+ * the coded video sequence.
+ */
+struct _GstH264SPSExtMVCLevelValue
+{
+  guint8 level_idc;
+  guint16 num_applicable_ops_minus1;
+  GstH264SPSExtMVCLevelValueOp *applicable_op;
+};
+
+/**
+ * GstH264SPSExtMVC:
+ * @num_views_minus1: plus 1 specifies the maximum number of coded
+ *   views in the coded video sequence
+ * @view: array of #GstH264SPSExtMVCView
+ * @num_level_values_signalled_minus1: plus 1 specifies the number of
+ *   level values signalled for the coded video sequence.
+ * @level_value: array of #GstH264SPSExtMVCLevelValue
+ *
+ * Represents the parsed seq_parameter_set_mvc_extension().
+ */
+struct _GstH264SPSExtMVC
+{
+  guint16 num_views_minus1;
+  GstH264SPSExtMVCView *view;
+  guint8 num_level_values_signalled_minus1;
+  GstH264SPSExtMVCLevelValue *level_value;
+};
+
 /**
  * GstH264SPS:
  * @id: The ID of the sequence parameter set
@@ -422,6 +618,8 @@ struct _GstH264SPS
   guint8 constraint_set1_flag;
   guint8 constraint_set2_flag;
   guint8 constraint_set3_flag;
+  guint8 constraint_set4_flag;
+  guint8 constraint_set5_flag;
   guint8 level_idc;
 
   guint8 chroma_format_idc;
@@ -477,6 +675,12 @@ struct _GstH264SPS
   gint crop_rect_x, crop_rect_y;
   gint fps_num, fps_den;
   gboolean valid;
+
+  /* Subset SPS extensions */
+  guint8 extension_type;
+  union {
+    GstH264SPSExtMVC mvc;
+  } extension;
 };
 
 /**
@@ -539,6 +743,8 @@ struct _GstH264RefPicListModification
     guint32 abs_diff_pic_num_minus1;
     /* if modification_of_pic_nums_idc == 2 */
     guint32 long_term_pic_num;
+    /* if modification_of_pic_nums_idc == 4 || 5 */
+    guint32 abs_diff_view_idx_minus1;
   } value;
 };
 
@@ -670,6 +876,36 @@ struct _GstH264ClockTimestamp
   guint32 time_offset;
 };
 
+struct _GstH264FramePacking
+{
+  guint32 frame_packing_id;
+  guint8 frame_packing_cancel_flag;
+  guint8 frame_packing_type; /* GstH264FramePackingType */
+  guint8 quincunx_sampling_flag;
+  guint8 content_interpretation_type;
+  guint8 spatial_flipping_flag;
+  guint8 frame0_flipped_flag;
+  guint8 field_views_flag;
+  guint8 current_frame_is_frame0_flag;
+  guint8 frame0_self_contained_flag;
+  guint8 frame1_self_contained_flag;
+  guint8 frame0_grid_position_x;
+  guint8 frame0_grid_position_y;
+  guint8 frame1_grid_position_x;
+  guint8 frame1_grid_position_y;
+  guint16 frame_packing_repetition_period;
+};
+
+struct _GstH264StereoVideoInfo
+{
+  guint8 field_views_flag;
+  guint8 top_field_is_left_view_flag;
+  guint8 current_frame_is_left_view_flag;
+  guint8 next_frame_is_second_view_flag;
+  guint8 left_view_self_contained_flag;
+  guint8 right_view_self_contained_flag;
+};
+
 struct _GstH264PicTiming
 {
   guint32 cpb_removal_delay;
@@ -696,6 +932,14 @@ struct _GstH264BufferingPeriod
   guint8 vcl_initial_cpb_removal_delay_offset[32];
 };
 
+struct _GstH264RecoveryPoint
+{
+  guint32 recovery_frame_cnt;
+  guint8 exact_match_flag;
+  guint8 broken_link_flag;
+  guint8 changing_slice_group_idc;
+};
+
 struct _GstH264SEIMessage
 {
   GstH264SEIPayloadType payloadType;
@@ -703,6 +947,9 @@ struct _GstH264SEIMessage
   union {
     GstH264BufferingPeriod buffering_period;
     GstH264PicTiming pic_timing;
+    GstH264RecoveryPoint recovery_point;
+    GstH264StereoVideoInfo stereo_video_info;
+    GstH264FramePacking frame_packing;
     /* ... could implement more */
   } payload;
 };
@@ -742,6 +989,9 @@ GstH264ParserResult gst_h264_parser_parse_slice_hdr   (GstH264NalParser *nalpars
                                                        GstH264SliceHdr *slice, gboolean parse_pred_weight_table,
                                                        gboolean parse_dec_ref_pic_marking);
 
+GstH264ParserResult gst_h264_parser_parse_subset_sps  (GstH264NalParser *nalparser, GstH264NalUnit *nalu,
+                                                       GstH264SPS *sps, gboolean parse_vui_params);
+
 GstH264ParserResult gst_h264_parser_parse_sps         (GstH264NalParser *nalparser, GstH264NalUnit *nalu,
                                                        GstH264SPS *sps, gboolean parse_vui_params);
 
@@ -749,15 +999,38 @@ GstH264ParserResult gst_h264_parser_parse_pps         (GstH264NalParser *nalpars
                                                        GstH264NalUnit *nalu, GstH264PPS *pps);
 
 GstH264ParserResult gst_h264_parser_parse_sei         (GstH264NalParser *nalparser,
-                                                       GstH264NalUnit *nalu, GstH264SEIMessage *sei);
+                                                       GstH264NalUnit *nalu, GArray ** messages);
 
 void gst_h264_nal_parser_free                         (GstH264NalParser *nalparser);
 
+GstH264ParserResult gst_h264_parse_subset_sps         (GstH264NalUnit *nalu,
+                                                       GstH264SPS *sps, gboolean parse_vui_params);
+
 GstH264ParserResult gst_h264_parse_sps                (GstH264NalUnit *nalu,
                                                        GstH264SPS *sps, gboolean parse_vui_params);
 
+void                gst_h264_sps_clear                (GstH264SPS *sps);
+
 GstH264ParserResult gst_h264_parse_pps                (GstH264NalParser *nalparser,
                                                        GstH264NalUnit *nalu, GstH264PPS *pps);
 
+void                gst_h264_pps_clear                (GstH264PPS *pps);
+
+void    gst_h264_quant_matrix_8x8_get_zigzag_from_raster (guint8 out_quant[64],
+                                                          const guint8 quant[64]);
+
+void    gst_h264_quant_matrix_8x8_get_raster_from_zigzag (guint8 out_quant[64],
+                                                          const guint8 quant[64]);
+
+void    gst_h264_quant_matrix_4x4_get_zigzag_from_raster (guint8 out_quant[16],
+                                                          const guint8 quant[16]);
+
+void    gst_h264_quant_matrix_4x4_get_raster_from_zigzag (guint8 out_quant[16],
+                                                          const guint8 quant[16]);
+
+void gst_h264_video_calculate_framerate (const GstH264SPS * sps, guint field_pic_flag,
+    guint pic_struct, gint * fps_num, gint * fps_den);
+
 G_END_DECLS
+
 #endif
