@@ -1,7 +1,7 @@
 /*
  *  gstvaapifilter.c - Video processing abstraction
  *
- *  Copyright (C) 2013 Intel Corporation
+ *  Copyright (C) 2013-2014 Intel Corporation
  *    Author: Gwenole Beauchesne <gwenole.beauchesne@intel.com>
  *
  *  This library is free software; you can redistribute it and/or
@@ -271,7 +271,7 @@ init_properties(void)
                           "The forced output pixel format",
                           GST_TYPE_VIDEO_FORMAT,
                           DEFAULT_FORMAT,
-                          G_PARAM_READWRITE);
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:crop-rect:
@@ -283,7 +283,7 @@ init_properties(void)
                            "Cropping Rectangle",
                            "The cropping rectangle",
                            GST_VAAPI_TYPE_RECTANGLE,
-                           G_PARAM_READWRITE);
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:denoise:
@@ -295,7 +295,7 @@ init_properties(void)
                            "Denoising Level",
                            "The level of denoising to apply",
                            0.0, 1.0, 0.0,
-                           G_PARAM_READWRITE);
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:sharpen:
@@ -308,7 +308,7 @@ init_properties(void)
                            "Sharpening Level",
                            "The level of sharpening/blurring to apply",
                            -1.0, 1.0, 0.0,
-                           G_PARAM_READWRITE);
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:hue:
@@ -321,7 +321,7 @@ init_properties(void)
                            "Hue",
                            "The color hue value",
                            -180.0, 180.0, 0.0,
-                           G_PARAM_READWRITE);
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:saturation:
@@ -334,7 +334,7 @@ init_properties(void)
                            "Saturation",
                            "The color saturation value",
                            0.0, 2.0, 1.0,
-                           G_PARAM_READWRITE);
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:brightness:
@@ -347,7 +347,7 @@ init_properties(void)
                            "Brightness",
                            "The color brightness value",
                            -1.0, 1.0, 0.0,
-                           G_PARAM_READWRITE);
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:contrast:
@@ -360,7 +360,7 @@ init_properties(void)
                            "Contrast",
                            "The color contrast value",
                            0.0, 2.0, 1.0,
-                           G_PARAM_READWRITE);
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
     /**
      * GstVaapiFilter:deinterlace-method:
@@ -374,7 +374,7 @@ init_properties(void)
                           "Deinterlacing method to apply",
                           GST_VAAPI_TYPE_DEINTERLACE_METHOD,
                           GST_VAAPI_DEINTERLACE_METHOD_NONE,
-                          G_PARAM_READWRITE);
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 }
 
 static void
@@ -483,14 +483,14 @@ op_data_ensure_caps(GstVaapiFilterOpData *op_data, gpointer filter_caps,
     guint num_filter_caps)
 {
     guchar *filter_cap = filter_caps;
-    guint i;
+    guint i, va_num_caps = num_filter_caps;
 
     // Find the VA filter cap matching the op info sub-type
     if (op_data->va_subtype) {
         for (i = 0; i < num_filter_caps; i++) {
             /* XXX: sub-type shall always be the first field */
             if (op_data->va_subtype == *(guint *)filter_cap) {
-                num_filter_caps = 1;
+                va_num_caps= 1;
                 break;
             }
             filter_cap += op_data->va_cap_size;
@@ -500,11 +500,11 @@ op_data_ensure_caps(GstVaapiFilterOpData *op_data, gpointer filter_caps,
     }
 
     op_data->va_caps = g_memdup(filter_cap,
-        op_data->va_cap_size * num_filter_caps);
+        op_data->va_cap_size * va_num_caps);
     if (!op_data->va_caps)
         return FALSE;
 
-    op_data->va_num_caps = num_filter_caps;
+    op_data->va_num_caps = va_num_caps;
     return TRUE;
 }
 
