@@ -3,7 +3,7 @@
  *
  *  Copyright (C) 2010-2011 Splitted-Desktop Systems
  *    Author: Gwenole Beauchesne <gwenole.beauchesne@splitted-desktop.com>
- *  Copyright (C) 2011-2013 Intel Corporation
+ *  Copyright (C) 2011-2014 Intel Corporation
  *    Author: Gwenole Beauchesne <gwenole.beauchesne@intel.com>
  *
  *  This library is free software; you can redistribute it and/or
@@ -35,6 +35,7 @@
 #include "gstvaapicontext.h"
 #include "gstvaapiimage.h"
 #include "gstvaapiimage_priv.h"
+#include "gstvaapicontext_overlay.h"
 
 #define DEBUG 1
 #include "gstvaapidebug.h"
@@ -187,7 +188,7 @@ gst_vaapi_surface_create_with_format(GstVaapiSurface *surface,
 
     /* ERRORS */
 error_unsupported_format:
-    GST_ERROR("unsupported format %u", gst_vaapi_video_format_to_string(format));
+    GST_ERROR("unsupported format %s", gst_vaapi_video_format_to_string(format));
     return FALSE;
 #else
     return FALSE;
