@@ -32,7 +32,6 @@
 #include "gst/vaapi/sysdeps.h"
 #include <gst/gst.h>
 #include <gst/video/video.h>
-#include <gst/video/videocontext.h>
 
 #include "gstvaapidownload.h"
 #include "gstvaapipluginutil.h"
@@ -75,9 +74,8 @@ struct _TransformSizeCache {
 
 struct _GstVaapiDownload {
     /*< private >*/
-    GstBaseTransform    parent_instance;
+    GstVaapiPluginBase  parent_instance;
 
-    GstVaapiDisplay    *display;
     GstCaps            *allowed_caps;
     TransformSizeCache  transform_size_cache[2];
     GstVaapiVideoPool  *images;
@@ -89,49 +87,14 @@ struct _GstVaapiDownload {
 
 struct _GstVaapiDownloadClass {
     /*< private >*/
-    GstBaseTransformClass parent_class;
+    GstVaapiPluginBaseClass parent_class;
 };
 
-/* GstImplementsInterface interface */
-static gboolean
-gst_vaapidownload_implements_interface_supported(
-    GstImplementsInterface *iface,
-    GType                   type
-)
-{
-    return (type == GST_TYPE_VIDEO_CONTEXT);
-}
-
-static void
-gst_vaapidownload_implements_iface_init(GstImplementsInterfaceClass *iface)
-{
-    iface->supported = gst_vaapidownload_implements_interface_supported;
-}
-
-/* GstVideoContext interface */
-static void
-gst_vaapidownload_set_video_context(GstVideoContext *context, const gchar *type,
-    const GValue *value)
-{
-  GstVaapiDownload *download = GST_VAAPIDOWNLOAD (context);
-  gst_vaapi_set_display (type, value, &download->display);
-}
-
-static void
-gst_video_context_interface_init(GstVideoContextInterface *iface)
-{
-    iface->set_context = gst_vaapidownload_set_video_context;
-}
-
-#define GstVideoContextClass GstVideoContextInterface
 G_DEFINE_TYPE_WITH_CODE(
     GstVaapiDownload,
     gst_vaapidownload,
     GST_TYPE_BASE_TRANSFORM,
-    G_IMPLEMENT_INTERFACE(GST_TYPE_IMPLEMENTS_INTERFACE,
-                          gst_vaapidownload_implements_iface_init);
-    G_IMPLEMENT_INTERFACE(GST_TYPE_VIDEO_CONTEXT,
-                          gst_video_context_interface_init))
+    GST_VAAPI_PLUGIN_BASE_INIT_INTERFACES)
 
 static gboolean
 gst_vaapidownload_start(GstBaseTransform *trans);
@@ -199,7 +162,7 @@ gst_vaapidownload_destroy(GstVaapiDownload *download)
     }
 
     gst_vaapi_video_pool_replace(&download->images, NULL);
-    gst_vaapi_display_replace(&download->display, NULL);
+    GST_VAAPI_PLUGIN_BASE_DISPLAY_REPLACE(download, NULL);
 }
 
 static void
@@ -207,6 +170,7 @@ gst_vaapidownload_finalize(GObject *object)
 {
     gst_vaapidownload_destroy(GST_VAAPIDOWNLOAD(object));
 
+    gst_vaapi_plugin_base_finalize(GST_VAAPI_PLUGIN_BASE(object));
     G_OBJECT_CLASS(gst_vaapidownload_parent_class)->finalize(object);
 }
 
@@ -221,6 +185,8 @@ gst_vaapidownload_class_init(GstVaapiDownloadClass *klass)
     GST_DEBUG_CATEGORY_INIT(gst_debug_vaapidownload,
                             GST_PLUGIN_NAME, 0, GST_PLUGIN_DESC);
 
+    gst_vaapi_plugin_base_class_init(GST_VAAPI_PLUGIN_BASE_CLASS(klass));
+
     object_class->finalize        = gst_vaapidownload_finalize;
     trans_class->start            = gst_vaapidownload_start;
     trans_class->stop             = gst_vaapidownload_stop;
@@ -250,7 +216,8 @@ gst_vaapidownload_init(GstVaapiDownload *download)
 {
     GstPad *sinkpad, *srcpad;
 
-    download->display           = NULL;
+    gst_vaapi_plugin_base_init(GST_VAAPI_PLUGIN_BASE(download), GST_CAT_DEFAULT);
+
     download->allowed_caps      = NULL;
     download->images            = NULL;
     download->images_reset      = FALSE;
@@ -272,8 +239,7 @@ gst_vaapidownload_init(GstVaapiDownload *download)
 static inline gboolean
 gst_vaapidownload_ensure_display(GstVaapiDownload *download)
 {
-    return gst_vaapi_ensure_display(download, GST_VAAPI_DISPLAY_TYPE_ANY,
-        &download->display);
+    return gst_vaapi_plugin_base_ensure_display(GST_VAAPI_PLUGIN_BASE(download));
 }
 
 static gboolean
@@ -281,6 +247,8 @@ gst_vaapidownload_start(GstBaseTransform *trans)
 {
     GstVaapiDownload * const download = GST_VAAPIDOWNLOAD(trans);
 
+    if (!gst_vaapi_plugin_base_open(GST_VAAPI_PLUGIN_BASE(trans)))
+        return FALSE;
     if (!gst_vaapidownload_ensure_display(download))
         return FALSE;
     return TRUE;
@@ -289,9 +257,7 @@ gst_vaapidownload_start(GstBaseTransform *trans)
 static gboolean
 gst_vaapidownload_stop(GstBaseTransform *trans)
 {
-    GstVaapiDownload * const download = GST_VAAPIDOWNLOAD(trans);
-
-    gst_vaapi_display_replace(&download->display, NULL);
+    gst_vaapi_plugin_base_close(GST_VAAPI_PLUGIN_BASE(trans));
     return TRUE;
 }
 
@@ -316,6 +282,7 @@ gst_vaapidownload_update_src_caps(GstVaapiDownload *download, GstBuffer *buffer)
     GstVaapiVideoMeta *meta;
     GstVaapiSurface *surface;
     GstVideoFormat format;
+    GstVideoInfo vi;
     GstPad *srcpad;
     GstCaps *in_caps, *out_caps;
 
@@ -331,20 +298,22 @@ gst_vaapidownload_update_src_caps(GstVaapiDownload *download, GstBuffer *buffer)
         return TRUE;
 
     in_caps = GST_BUFFER_CAPS(buffer);
-    if (!in_caps) {
+    if (!in_caps || !gst_caps_is_fixed(in_caps)) {
         GST_WARNING("failed to retrieve caps from buffer");
         return FALSE;
     }
 
-    out_caps = gst_vaapi_video_format_to_caps(download->image_format);
-    if (!out_caps) {
-        GST_WARNING("failed to create caps from format %s",
-                    gst_video_format_to_string(download->image_format));
+    if (!gst_video_info_from_caps(&vi, in_caps)) {
+        GST_WARNING("failed to parse caps %" GST_PTR_FORMAT, in_caps);
         return FALSE;
     }
 
-    if (!gst_vaapi_append_surface_caps(out_caps, in_caps)) {
-        gst_caps_unref(out_caps);
+    gst_video_info_set_format(&vi, download->image_format,
+        GST_VIDEO_INFO_WIDTH(&vi), GST_VIDEO_INFO_HEIGHT(&vi));
+    out_caps = gst_video_info_to_caps(&vi);
+    if (!out_caps) {
+        GST_WARNING("failed to create caps from format %s",
+                    gst_video_format_to_string(download->image_format));
         return FALSE;
     }
 
@@ -423,8 +392,9 @@ gst_vaapidownload_transform_caps(
 {
     GstVaapiDownload * const download = GST_VAAPIDOWNLOAD(trans);
     GstPad *srcpad;
-    GstCaps *allowed_caps, *inter_caps, *out_caps = NULL;
+    GstCaps *allowed_caps, *tmp_caps, *out_caps = NULL;
     GstStructure *structure;
+    GArray *formats;
 
     g_return_val_if_fail(GST_IS_CAPS(caps), NULL);
 
@@ -443,23 +413,31 @@ gst_vaapidownload_transform_caps(
         if (download->allowed_caps)
             allowed_caps = gst_caps_ref(download->allowed_caps);
         else {
-            allowed_caps = gst_vaapi_display_get_image_caps(download->display);
+            formats = gst_vaapi_display_get_image_formats(
+                GST_VAAPI_PLUGIN_BASE_DISPLAY(download));
+            if (G_UNLIKELY(!formats))
+                allowed_caps = gst_caps_new_empty();
+            else {
+                allowed_caps =
+                    gst_vaapi_video_format_new_template_caps_from_list(formats);
+                g_array_unref(formats);
+            }
             if (!allowed_caps)
                 return NULL;
         }
-        inter_caps = gst_caps_intersect(out_caps, allowed_caps);
+        tmp_caps = gst_caps_intersect(out_caps, allowed_caps);
         gst_caps_unref(allowed_caps);
         gst_caps_unref(out_caps);
-        out_caps = inter_caps;
+        out_caps = tmp_caps;
 
         /* Intersect with allowed caps from the peer, if any */
         srcpad = gst_element_get_static_pad(GST_ELEMENT(download), "src");
         allowed_caps = gst_pad_peer_get_caps(srcpad);
         if (allowed_caps) {
-            inter_caps = gst_caps_intersect(out_caps, allowed_caps);
+            tmp_caps = gst_caps_intersect(out_caps, allowed_caps);
             gst_caps_unref(allowed_caps);
             gst_caps_unref(out_caps);
-            out_caps = inter_caps;
+            out_caps = tmp_caps;
         }
     }
     else {
@@ -504,7 +482,8 @@ gst_vaapidownload_ensure_image_pool(GstVaapiDownload *download, GstCaps *caps)
         download->image_width  = width;
         download->image_height = height;
         gst_vaapi_video_pool_replace(&download->images, NULL);
-        download->images = gst_vaapi_image_pool_new(download->display, &vi);
+        download->images = gst_vaapi_image_pool_new(
+            GST_VAAPI_PLUGIN_BASE_DISPLAY(download), &vi);
         if (!download->images)
             return FALSE;
         download->images_reset = TRUE;
@@ -531,8 +510,11 @@ gst_vaapidownload_set_caps(
     GstCaps          *outcaps
 )
 {
+    GstVaapiPluginBase * const plugin = GST_VAAPI_PLUGIN_BASE(trans);
     GstVaapiDownload * const download = GST_VAAPIDOWNLOAD(trans);
 
+    if (!gst_vaapi_plugin_base_set_caps(plugin, incaps, outcaps))
+        return FALSE;
     if (!gst_vaapidownload_negotiate_buffers(download, incaps, outcaps))
         return FALSE;
     return TRUE;
@@ -587,11 +569,12 @@ static gboolean
 gst_vaapidownload_query(GstPad *pad, GstQuery *query)
 {
     GstVaapiDownload * const download = GST_VAAPIDOWNLOAD(gst_pad_get_parent_element(pad));
+    GstVaapiDisplay * const display = GST_VAAPI_PLUGIN_BASE_DISPLAY(download);
     gboolean res;
 
-    GST_DEBUG("sharing display %p", download->display);
+    GST_DEBUG("sharing display %p", display);
 
-    if (gst_vaapi_reply_to_query(query, download->display))
+    if (gst_vaapi_reply_to_query(query, display))
         res = TRUE;
     else
         res = gst_pad_query_default(pad, query);
