@@ -1,7 +1,7 @@
 /*
  *  gstvaapipostproc.c - VA-API video postprocessing
  *
- *  Copyright (C) 2012-2013 Intel Corporation
+ *  Copyright (C) 2012-2014 Intel Corporation
  *    Author: Gwenole Beauchesne <gwenole.beauchesne@intel.com>
  *
  *  This library is free software; you can redistribute it and/or
@@ -34,7 +34,6 @@
 
 #include "gstvaapipostproc.h"
 #include "gstvaapipluginutil.h"
-#include "gstvaapivideocontext.h"
 #include "gstvaapivideobuffer.h"
 #if GST_CHECK_VERSION(1,0,0)
 #include "gstvaapivideobufferpool.h"
@@ -89,63 +88,11 @@ static GstStaticPadTemplate gst_vaapipostproc_src_factory =
         GST_PAD_ALWAYS,
         GST_STATIC_CAPS(gst_vaapipostproc_src_caps_str));
 
-/* GstImplementsInterface interface */
-#if !GST_CHECK_VERSION(1,0,0)
-static gboolean
-gst_vaapipostproc_implements_interface_supported(
-    GstImplementsInterface *iface,
-    GType                   type
-)
-{
-    return (type == GST_TYPE_VIDEO_CONTEXT);
-}
-
-static void
-gst_vaapipostproc_implements_iface_init(GstImplementsInterfaceClass *iface)
-{
-    iface->supported = gst_vaapipostproc_implements_interface_supported;
-}
-#endif
-
-/* GstVideoContext interface */
-#if !GST_CHECK_VERSION(1,1,0)
-static void
-gst_vaapipostproc_set_video_context(
-    GstVideoContext *context,
-    const gchar     *type,
-    const GValue    *value
-)
-{
-    GstVaapiPostproc * const postproc = GST_VAAPIPOSTPROC(context);
-
-    gst_vaapi_set_display(type, value, &postproc->display);
-
-    if (postproc->uploader)
-        gst_vaapi_uploader_ensure_display(postproc->uploader, postproc->display);
-}
-
-static void
-gst_video_context_interface_init(GstVideoContextInterface *iface)
-{
-    iface->set_context = gst_vaapipostproc_set_video_context;
-}
-
-#define GstVideoContextClass GstVideoContextInterface
-#endif
-
 G_DEFINE_TYPE_WITH_CODE(
     GstVaapiPostproc,
     gst_vaapipostproc,
     GST_TYPE_BASE_TRANSFORM,
-#if !GST_CHECK_VERSION(1,0,0)
-    G_IMPLEMENT_INTERFACE(GST_TYPE_IMPLEMENTS_INTERFACE,
-                          gst_vaapipostproc_implements_iface_init);
-#endif
-#if !GST_CHECK_VERSION(1,1,0)
-    G_IMPLEMENT_INTERFACE(GST_TYPE_VIDEO_CONTEXT,
-                          gst_video_context_interface_init)
-#endif
-    )
+    GST_VAAPI_PLUGIN_BASE_INIT_INTERFACES)
 
 enum {
     PROP_0,
@@ -158,6 +105,10 @@ enum {
     PROP_DEINTERLACE_METHOD,
     PROP_DENOISE,
     PROP_SHARPEN,
+    PROP_HUE,
+    PROP_SATURATION,
+    PROP_BRIGHTNESS,
+    PROP_CONTRAST,
 };
 
 #define DEFAULT_FORMAT                  GST_VIDEO_FORMAT_ENCODED
@@ -254,25 +205,10 @@ find_filter_op(GPtrArray *filter_ops, GstVaapiFilterOp op)
     return NULL;
 }
 
-#if GST_CHECK_VERSION(1,1,0)
-static void
-gst_vaapipostproc_set_context(GstElement *element, GstContext *context)
-{
-    GstVaapiPostproc * const postproc = GST_VAAPIPOSTPROC(element);
-    GstVaapiDisplay *display = NULL;
-
-    if (gst_vaapi_video_context_get_display(context, &display)) {
-        GST_INFO_OBJECT(element, "set display %p", display);
-        gst_vaapi_display_replace(&postproc->display, display);
-    }
-}
-#endif
-
 static inline gboolean
 gst_vaapipostproc_ensure_display(GstVaapiPostproc *postproc)
 {
-    return gst_vaapi_ensure_display(postproc, GST_VAAPI_DISPLAY_TYPE_ANY,
-        &postproc->display);
+    return gst_vaapi_plugin_base_ensure_display(GST_VAAPI_PLUGIN_BASE(postproc));
 }
 
 static gboolean
@@ -280,30 +216,11 @@ gst_vaapipostproc_ensure_uploader(GstVaapiPostproc *postproc)
 {
     if (!gst_vaapipostproc_ensure_display(postproc))
         return FALSE;
-
-    if (!postproc->uploader) {
-        postproc->uploader = gst_vaapi_uploader_new(postproc->display);
-        if (!postproc->uploader)
-            return FALSE;
-    }
-
-    if (!gst_vaapi_uploader_ensure_display(postproc->uploader,
-            postproc->display))
+    if (!gst_vaapi_plugin_base_ensure_uploader(GST_VAAPI_PLUGIN_BASE(postproc)))
         return FALSE;
     return TRUE;
 }
 
-static gboolean
-gst_vaapipostproc_ensure_uploader_caps(GstVaapiPostproc *postproc)
-{
-#if !GST_CHECK_VERSION(1,0,0)
-    if (postproc->is_raw_yuv && !gst_vaapi_uploader_ensure_caps(
-            postproc->uploader, postproc->sinkpad_caps, NULL))
-        return FALSE;
-#endif
-    return TRUE;
-}
-
 static gboolean
 gst_vaapipostproc_ensure_filter(GstVaapiPostproc *postproc)
 {
@@ -313,7 +230,8 @@ gst_vaapipostproc_ensure_filter(GstVaapiPostproc *postproc)
     if (!gst_vaapipostproc_ensure_display(postproc))
         return FALSE;
 
-    postproc->filter = gst_vaapi_filter_new(postproc->display);
+    postproc->filter = gst_vaapi_filter_new(
+        GST_VAAPI_PLUGIN_BASE_DISPLAY(postproc));
     if (!postproc->filter)
         return FALSE;
     return TRUE;
@@ -338,12 +256,12 @@ gst_vaapipostproc_ensure_filter_caps(GstVaapiPostproc *postproc)
 static gboolean
 gst_vaapipostproc_create(GstVaapiPostproc *postproc)
 {
+    if (!gst_vaapi_plugin_base_open(GST_VAAPI_PLUGIN_BASE(postproc)))
+        return FALSE;
     if (!gst_vaapipostproc_ensure_display(postproc))
         return FALSE;
     if (!gst_vaapipostproc_ensure_uploader(postproc))
         return FALSE;
-    if (!gst_vaapipostproc_ensure_uploader_caps(postproc))
-        return FALSE;
     if (gst_vaapipostproc_ensure_filter(postproc))
         postproc->use_vpp = TRUE;
     return TRUE;
@@ -369,17 +287,11 @@ static void
 gst_vaapipostproc_destroy(GstVaapiPostproc *postproc)
 {
     ds_reset(&postproc->deinterlace_state);
-#if GST_CHECK_VERSION(1,0,0)
-    g_clear_object(&postproc->sinkpad_buffer_pool);
-#endif
-    g_clear_object(&postproc->uploader);
     gst_vaapipostproc_destroy_filter(postproc);
-    gst_vaapi_display_replace(&postproc->display, NULL);
 
     gst_caps_replace(&postproc->allowed_sinkpad_caps, NULL);
-    gst_caps_replace(&postproc->sinkpad_caps, NULL);
     gst_caps_replace(&postproc->allowed_srcpad_caps, NULL);
-    gst_caps_replace(&postproc->srcpad_caps,  NULL);
+    gst_vaapi_plugin_base_close(GST_VAAPI_PLUGIN_BASE(postproc));
 }
 
 static gboolean
@@ -388,6 +300,8 @@ gst_vaapipostproc_start(GstBaseTransform *trans)
     GstVaapiPostproc * const postproc = GST_VAAPIPOSTPROC(trans);
 
     ds_reset(&postproc->deinterlace_state);
+    if (!gst_vaapi_plugin_base_open(GST_VAAPI_PLUGIN_BASE(postproc)))
+        return FALSE;
     if (!gst_vaapipostproc_ensure_display(postproc))
         return FALSE;
     return TRUE;
@@ -399,30 +313,41 @@ gst_vaapipostproc_stop(GstBaseTransform *trans)
     GstVaapiPostproc * const postproc = GST_VAAPIPOSTPROC(trans);
 
     ds_reset(&postproc->deinterlace_state);
-    gst_vaapi_display_replace(&postproc->display, NULL);
+    gst_vaapi_plugin_base_close(GST_VAAPI_PLUGIN_BASE(postproc));
     return TRUE;
 }
 
 static gboolean
-is_interlaced_buffer(GstVaapiPostproc *postproc, GstBuffer *buf)
+should_deinterlace_buffer(GstVaapiPostproc *postproc, GstBuffer *buf)
 {
-    if (!(postproc->flags & GST_VAAPI_POSTPROC_FLAG_DEINTERLACE))
+    if (!(postproc->flags & GST_VAAPI_POSTPROC_FLAG_DEINTERLACE) ||
+        postproc->deinterlace_mode == GST_VAAPI_DEINTERLACE_MODE_DISABLED)
         return FALSE;
 
+    if (postproc->deinterlace_mode == GST_VAAPI_DEINTERLACE_MODE_INTERLACED)
+        return TRUE;
+
+    g_assert(postproc->deinterlace_mode == GST_VAAPI_DEINTERLACE_MODE_AUTO);
+
     switch (GST_VIDEO_INFO_INTERLACE_MODE(&postproc->sinkpad_info)) {
+    case GST_VIDEO_INTERLACE_MODE_INTERLEAVED:
+        return TRUE;
+    case GST_VIDEO_INTERLACE_MODE_PROGRESSIVE:
+        return FALSE;
     case GST_VIDEO_INTERLACE_MODE_MIXED:
 #if GST_CHECK_VERSION(1,0,0)
-        if (!GST_BUFFER_FLAG_IS_SET(buf, GST_VIDEO_BUFFER_FLAG_INTERLACED))
-            return FALSE;
+        if (GST_BUFFER_FLAG_IS_SET(buf, GST_VIDEO_BUFFER_FLAG_INTERLACED))
+            return TRUE;
 #else
-        if (GST_BUFFER_FLAG_IS_SET(buf, GST_VIDEO_BUFFER_PROGRESSIVE))
-            return FALSE;
+        if (!GST_BUFFER_FLAG_IS_SET(buf, GST_VIDEO_BUFFER_PROGRESSIVE))
+            return TRUE;
 #endif
         break;
     default:
+        GST_ERROR("unhandled \"interlace-mode\", disabling deinterlacing" );
         break;
     }
-    return TRUE;
+    return FALSE;
 }
 
 static GstBuffer *
@@ -437,7 +362,7 @@ create_output_buffer(GstVaapiPostproc *postproc)
         goto error_create_buffer;
 
 #if !GST_CHECK_VERSION(1,0,0)
-    gst_buffer_set_caps(outbuf, postproc->srcpad_caps);
+    gst_buffer_set_caps(outbuf, GST_VAAPI_PLUGIN_BASE_SRC_PAD_CAPS(postproc));
 #endif
     return outbuf;
 
@@ -522,6 +447,10 @@ gst_vaapipostproc_process_vpp(GstBaseTransform *trans, GstBuffer *inbuf,
     GstVaapiDeinterlaceMethod deint_method;
     guint flags, deint_flags;
     gboolean tff, deint, deint_refs, deint_changed;
+    GstVaapiRectangle *crop_rect = NULL;
+#if GST_CHECK_VERSION(1,0,0)
+    GstVaapiRectangle tmp_rect;
+#endif
 
     /* Validate filters */
     if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_FORMAT) &&
@@ -534,33 +463,80 @@ gst_vaapipostproc_process_vpp(GstBaseTransform *trans, GstBuffer *inbuf,
         return GST_FLOW_NOT_SUPPORTED;
 
     if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_SHARPEN) &&
-        !gst_vaapi_filter_set_denoising_level(postproc->filter,
+        !gst_vaapi_filter_set_sharpening_level(postproc->filter,
             postproc->sharpen_level))
         return GST_FLOW_NOT_SUPPORTED;
 
+    if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_HUE) &&
+        !gst_vaapi_filter_set_hue(postproc->filter,
+            postproc->hue))
+        return GST_FLOW_NOT_SUPPORTED;
+
+    if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_SATURATION) &&
+        !gst_vaapi_filter_set_saturation(postproc->filter,
+            postproc->saturation))
+        return GST_FLOW_NOT_SUPPORTED;
+
+    if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_BRIGHTNESS) &&
+        !gst_vaapi_filter_set_brightness(postproc->filter,
+            postproc->brightness))
+        return GST_FLOW_NOT_SUPPORTED;
+
+    if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_CONTRAST) &&
+        !gst_vaapi_filter_set_contrast(postproc->filter,
+            postproc->contrast))
+        return GST_FLOW_NOT_SUPPORTED;
+
     inbuf_meta = gst_buffer_get_vaapi_video_meta(inbuf);
     if (!inbuf_meta)
         goto error_invalid_buffer;
     inbuf_surface = gst_vaapi_video_meta_get_surface(inbuf_meta);
 
+#if GST_CHECK_VERSION(1,0,0)
+    GstVideoCropMeta * const crop_meta =
+        gst_buffer_get_video_crop_meta(inbuf);
+    if (crop_meta) {
+        crop_rect = &tmp_rect;
+        crop_rect->x = crop_meta->x;
+        crop_rect->y = crop_meta->y;
+        crop_rect->width = crop_meta->width;
+        crop_rect->height = crop_meta->height;
+    }
+#endif
+    if (!crop_rect)
+        crop_rect = (GstVaapiRectangle *)
+            gst_vaapi_video_meta_get_render_rect(inbuf_meta);
+
     timestamp  = GST_BUFFER_TIMESTAMP(inbuf);
     tff        = GST_BUFFER_FLAG_IS_SET(inbuf, GST_VIDEO_BUFFER_FLAG_TFF);
-    deint      = is_interlaced_buffer(postproc, inbuf);
+    deint      = should_deinterlace_buffer(postproc, inbuf);
 
     /* Drop references if deinterlacing conditions changed */
     deint_changed = deint != ds->deint;
     if (deint_changed || (ds->num_surfaces > 0 && tff != ds->tff))
         ds_reset(ds);
+
+    deint_method = postproc->deinterlace_method;
+    deint_refs = deint_method_is_advanced(deint_method);
+    if (deint_refs) {
+        GstBuffer * const prev_buf = ds_get_buffer(ds, 0);
+        GstClockTime prev_pts, pts = GST_BUFFER_TIMESTAMP(inbuf);
+        /* Reset deinterlacing state when there is a discontinuity */
+        if (prev_buf && (prev_pts = GST_BUFFER_TIMESTAMP(prev_buf)) != pts) {
+            const GstClockTimeDiff pts_diff = GST_CLOCK_DIFF(prev_pts, pts);
+            if (pts_diff < 0 || (postproc->field_duration > 0 &&
+                    pts_diff > postproc->field_duration * 2))
+                ds_reset(ds);
+        }
+    }
+
     ds->deint = deint;
     ds->tff = tff;
 
     flags = gst_vaapi_video_meta_get_render_flags(inbuf_meta) &
-        ~(GST_VAAPI_PICTURE_STRUCTURE_TOP_FIELD|
-          GST_VAAPI_PICTURE_STRUCTURE_BOTTOM_FIELD);
+        ~GST_VAAPI_PICTURE_STRUCTURE_MASK;
 
     /* First field */
-    deint_method = postproc->deinterlace_method;
-    deint_refs = deint_method_is_advanced(deint_method);
     if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_DEINTERLACE) {
         fieldbuf = create_output_buffer(postproc);
         if (!fieldbuf)
@@ -601,6 +577,7 @@ gst_vaapipostproc_process_vpp(GstBaseTransform *trans, GstBuffer *inbuf,
                 goto error_op_deinterlace;
         }
 
+        gst_vaapi_filter_set_cropping_rectangle(postproc->filter, crop_rect);
         status = gst_vaapi_filter_process(postproc->filter, inbuf_surface,
             outbuf_surface, flags);
         if (status != GST_VAAPI_FILTER_STATUS_SUCCESS)
@@ -639,6 +616,7 @@ gst_vaapipostproc_process_vpp(GstBaseTransform *trans, GstBuffer *inbuf,
                  postproc->filter, deint_method, 0))
         goto error_op_deinterlace;
 
+    gst_vaapi_filter_set_cropping_rectangle(postproc->filter, crop_rect);
     status = gst_vaapi_filter_process(postproc->filter, inbuf_surface,
         outbuf_surface, flags);
     if (status != GST_VAAPI_FILTER_STATUS_SUCCESS)
@@ -715,11 +693,10 @@ gst_vaapipostproc_process(GstBaseTransform *trans, GstBuffer *inbuf,
 
     timestamp  = GST_BUFFER_TIMESTAMP(inbuf);
     tff        = GST_BUFFER_FLAG_IS_SET(inbuf, GST_VIDEO_BUFFER_FLAG_TFF);
-    deint      = is_interlaced_buffer(postproc, inbuf);
+    deint      = should_deinterlace_buffer(postproc, inbuf);
 
     flags = gst_vaapi_video_meta_get_render_flags(meta) &
-        ~(GST_VAAPI_PICTURE_STRUCTURE_TOP_FIELD|
-          GST_VAAPI_PICTURE_STRUCTURE_BOTTOM_FIELD);
+        ~GST_VAAPI_PICTURE_STRUCTURE_MASK;
 
     /* First field */
     fieldbuf = create_output_buffer(postproc);
@@ -849,21 +826,11 @@ gst_vaapipostproc_update_sink_caps(GstVaapiPostproc *postproc, GstCaps *caps,
     deinterlace = is_deinterlace_enabled(postproc, &vi);
     if (deinterlace)
         postproc->flags |= GST_VAAPI_POSTPROC_FLAG_DEINTERLACE;
-    postproc->field_duration = gst_util_uint64_scale(
-        GST_SECOND, GST_VIDEO_INFO_FPS_D(&vi),
-        (1 + deinterlace) * GST_VIDEO_INFO_FPS_N(&vi));
+    postproc->field_duration = GST_VIDEO_INFO_FPS_N(&vi) > 0 ?
+        gst_util_uint64_scale(GST_SECOND, GST_VIDEO_INFO_FPS_D(&vi),
+            (1 + deinterlace) * GST_VIDEO_INFO_FPS_N(&vi)) : 0;
 
     postproc->is_raw_yuv = GST_VIDEO_INFO_IS_YUV(&vi);
-#if !GST_CHECK_VERSION(1,0,0)
-    if (postproc->is_raw_yuv) {
-        /* Ensure the uploader is set up for upstream allocated buffers */
-        GstVaapiUploader * const uploader = postproc->uploader;
-        if (!gst_vaapi_uploader_ensure_display(uploader, postproc->display))
-            return FALSE;
-        if (!gst_vaapi_uploader_ensure_caps(uploader, caps, NULL))
-            return FALSE;
-    }
-#endif
     return TRUE;
 }
 
@@ -912,7 +879,7 @@ ensure_allowed_sinkpad_caps(GstVaapiPostproc *postproc)
 
     /* Append YUV caps */
     if (gst_vaapipostproc_ensure_uploader(postproc)) {
-        yuv_caps = gst_vaapi_uploader_get_caps(postproc->uploader);
+        yuv_caps = GST_VAAPI_PLUGIN_BASE_UPLOADER_CAPS(postproc);
         if (yuv_caps) {
             out_caps = gst_caps_make_writable(out_caps);
             gst_caps_append(out_caps, gst_caps_copy(yuv_caps));
@@ -926,57 +893,11 @@ ensure_allowed_sinkpad_caps(GstVaapiPostproc *postproc)
     return TRUE;
 }
 
-/* Build list of supported formats */
-static gboolean
-build_format_list_value(GArray *formats, GValue *out_value)
-{
-    GValue v_format = { 0, };
-    guint i;
-#if GST_CHECK_VERSION(1,0,0)
-    const gchar *str;
-
-    g_value_init(out_value, GST_TYPE_LIST);
-
-    g_value_init(&v_format, G_TYPE_STRING);
-    g_value_set_string(&v_format, "encoded");
-    gst_value_list_append_value(out_value, &v_format);
-
-    for (i = 0; i < formats->len; i++) {
-        GstVideoFormat const format =
-            g_array_index(formats, GstVideoFormat, i);
-
-        str = gst_vaapi_video_format_to_string(format);
-        if (!str)
-            continue;
-        g_value_set_string(&v_format, str);
-        gst_value_list_append_value(out_value, &v_format);
-    }
-#else
-    guint32 fourcc;
-
-    g_value_init(out_value, GST_TYPE_LIST);
-    g_value_init(&v_format, GST_TYPE_FOURCC);
-    for (i = 0; i < formats->len; i++) {
-        GstVideoFormat const format =
-            g_array_index(formats, GstVideoFormat, i);
-
-        fourcc = gst_video_format_to_fourcc(format);
-        if (!fourcc)
-            continue;
-        gst_value_set_fourcc(&v_format, fourcc);
-        gst_value_list_append_value(out_value, &v_format);
-    }
-#endif
-
-    g_value_unset(&v_format);
-    return TRUE;
-}
-
 /* Fixup output caps so that to reflect the supported set of pixel formats */
 static GstCaps *
 expand_allowed_srcpad_caps(GstVaapiPostproc *postproc, GstCaps *caps)
 {
-    GValue value = { 0, };
+    GValue value = G_VALUE_INIT, v_format = G_VALUE_INIT;
     guint i, num_structures;
     gboolean had_filter;
 
@@ -987,8 +908,12 @@ expand_allowed_srcpad_caps(GstVaapiPostproc *postproc, GstCaps *caps)
         goto cleanup;
 
     /* Reset "format" field for each structure */
-    if (!build_format_list_value(postproc->filter_formats, &value))
+    if (!gst_vaapi_value_set_format_list(&value, postproc->filter_formats))
         goto cleanup;
+    if (gst_vaapi_value_set_format(&v_format, GST_VIDEO_FORMAT_ENCODED)) {
+        gst_value_list_prepend_value(&value, &v_format);
+        g_value_unset(&v_format);
+    }
 
     num_structures = gst_caps_get_size(caps);
     for (i = 0; i < num_structures; i++) {
@@ -1180,107 +1105,6 @@ gst_vaapipostproc_transform_size(GstBaseTransform *trans,
     return TRUE;
 }
 
-static GstBuffer *
-get_source_buffer(GstVaapiPostproc *postproc, GstBuffer *inbuf)
-{
-    GstVaapiVideoMeta *meta;
-    GstBuffer *outbuf;
-#if GST_CHECK_VERSION(1,0,0)
-    GstVideoFrame src_frame, out_frame;
-#endif
-
-    meta = gst_buffer_get_vaapi_video_meta(inbuf);
-    if (meta)
-        return gst_buffer_ref(inbuf);
-
-#if GST_CHECK_VERSION(1,0,0)
-    if (!postproc->is_raw_yuv)
-        goto error_invalid_buffer;
-
-    if (!postproc->sinkpad_buffer_pool)
-        goto error_no_pool;
-
-    if (!gst_buffer_pool_set_active(postproc->sinkpad_buffer_pool, TRUE))
-        goto error_active_pool;
-
-    outbuf = NULL;
-    if (gst_buffer_pool_acquire_buffer(postproc->sinkpad_buffer_pool,
-            &outbuf, NULL) != GST_FLOW_OK)
-        goto error_create_buffer;
-
-    if (!gst_video_frame_map(&src_frame, &postproc->sinkpad_info, inbuf,
-            GST_MAP_READ))
-        goto error_map_src_buffer;
-
-    if (!gst_video_frame_map(&out_frame, &postproc->sinkpad_info, outbuf,
-            GST_MAP_WRITE))
-        goto error_map_dst_buffer;
-
-    if (!gst_video_frame_copy(&out_frame, &src_frame))
-        goto error_copy_buffer;
-
-    gst_video_frame_unmap(&out_frame);
-    gst_video_frame_unmap(&src_frame);
-    gst_buffer_copy_into(outbuf, inbuf, GST_BUFFER_COPY_TIMESTAMPS, 0, -1);
-    return outbuf;
-
-    /* ERRORS */
-error_invalid_buffer:
-    {
-        GST_ERROR("failed to validate source buffer");
-        return NULL;
-    }
-error_no_pool:
-    {
-        GST_ERROR("no buffer pool was negotiated");
-        return NULL;
-    }
-error_active_pool:
-    {
-        GST_ERROR("failed to activate buffer pool");
-        return NULL;
-    }
-error_map_dst_buffer:
-    {
-        gst_video_frame_unmap(&src_frame);
-        // fall-through
-    }
-error_map_src_buffer:
-    {
-        GST_ERROR("failed to map buffer");
-        gst_buffer_unref(outbuf);
-        return NULL;
-    }
-#else
-    outbuf = gst_vaapi_uploader_get_buffer(postproc->uploader);
-    if (!outbuf)
-        goto error_create_buffer;
-    if (!gst_vaapi_uploader_process(postproc->uploader, inbuf, outbuf))
-        goto error_copy_buffer;
-
-    gst_buffer_copy_metadata(outbuf, inbuf,
-        GST_BUFFER_COPY_FLAGS | GST_BUFFER_COPY_TIMESTAMPS);
-    return outbuf;
-#endif
-
-    /* ERRORS */
-error_create_buffer:
-    {
-        GST_ERROR("failed to create buffer");
-        return NULL;
-    }
-error_copy_buffer:
-    {
-        GST_ERROR("failed to upload buffer to VA surface");
-#if GST_CHECK_VERSION(1,0,0)
-        gst_video_frame_unmap(&out_frame);
-        gst_video_frame_unmap(&src_frame);
-#endif
-        gst_buffer_unref(outbuf);
-        return NULL;
-    }
-}
-
 static GstFlowReturn
 gst_vaapipostproc_transform(GstBaseTransform *trans, GstBuffer *inbuf,
     GstBuffer *outbuf)
@@ -1289,8 +1113,9 @@ gst_vaapipostproc_transform(GstBaseTransform *trans, GstBuffer *inbuf,
     GstBuffer *buf;
     GstFlowReturn ret;
 
-    buf = get_source_buffer(postproc, inbuf);
-    if (!buf)
+    ret = gst_vaapi_plugin_base_get_input_buffer(
+        GST_VAAPI_PLUGIN_BASE(postproc), inbuf, &buf);
+    if (ret != GST_FLOW_OK)
         return GST_FLOW_ERROR;
 
     ret = GST_FLOW_NOT_SUPPORTED;
@@ -1335,72 +1160,6 @@ gst_vaapipostproc_prepare_output_buffer(GstBaseTransform *trans,
     return *outbuf_ptr ? GST_FLOW_OK : GST_FLOW_ERROR;
 }
 
-static gboolean
-ensure_sinkpad_buffer_pool(GstVaapiPostproc *postproc, GstCaps *caps)
-{
-#if GST_CHECK_VERSION(1,0,0)
-    GstBufferPool *pool;
-    GstCaps *pool_caps;
-    GstStructure *config;
-    GstVideoInfo vi;
-    gboolean need_pool;
-
-    if (!gst_vaapipostproc_ensure_display(postproc))
-        return FALSE;
-
-    if (postproc->sinkpad_buffer_pool) {
-        config = gst_buffer_pool_get_config(postproc->sinkpad_buffer_pool);
-        gst_buffer_pool_config_get_params(config, &pool_caps, NULL, NULL, NULL);
-        need_pool = !gst_caps_is_equal(caps, pool_caps);
-        gst_structure_free(config);
-        if (!need_pool)
-            return TRUE;
-        g_clear_object(&postproc->sinkpad_buffer_pool);
-        postproc->sinkpad_buffer_size = 0;
-    }
-
-    pool = gst_vaapi_video_buffer_pool_new(postproc->display);
-    if (!pool)
-        goto error_create_pool;
-
-    gst_video_info_init(&vi);
-    gst_video_info_from_caps(&vi, caps);
-    if (GST_VIDEO_INFO_FORMAT(&vi) == GST_VIDEO_FORMAT_ENCODED) {
-        GST_DEBUG("assume sink pad buffer pool format is NV12");
-        gst_video_info_set_format(&vi, GST_VIDEO_FORMAT_NV12,
-            GST_VIDEO_INFO_WIDTH(&vi), GST_VIDEO_INFO_HEIGHT(&vi));
-    }
-    postproc->sinkpad_buffer_size = vi.size;
-
-    config = gst_buffer_pool_get_config(pool);
-    gst_buffer_pool_config_set_params(config, caps,
-        postproc->sinkpad_buffer_size, 0, 0);
-    gst_buffer_pool_config_add_option(config,
-        GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META);
-    gst_buffer_pool_config_add_option(config,
-        GST_BUFFER_POOL_OPTION_VIDEO_META);
-    if (!gst_buffer_pool_set_config(pool, config))
-        goto error_pool_config;
-    postproc->sinkpad_buffer_pool = pool;
-    return TRUE;
-
-    /* ERRORS */
-error_create_pool:
-    {
-        GST_ERROR("failed to create buffer pool");
-        return FALSE;
-    }
-error_pool_config:
-    {
-        GST_ERROR("failed to reset buffer pool config");
-        gst_object_unref(pool);
-        return FALSE;
-    }
-#else
-    return TRUE;
-#endif
-}
-
 static gboolean
 ensure_srcpad_buffer_pool(GstVaapiPostproc *postproc, GstCaps *caps)
 {
@@ -1412,11 +1171,11 @@ ensure_srcpad_buffer_pool(GstVaapiPostproc *postproc, GstCaps *caps)
     gst_video_info_set_format(&vi, postproc->format,
         GST_VIDEO_INFO_WIDTH(&vi), GST_VIDEO_INFO_HEIGHT(&vi));
 
-    if (!video_info_changed(&vi, &postproc->filter_pool_info))
+    if (postproc->filter_pool && !video_info_changed(&vi, &postproc->filter_pool_info))
         return TRUE;
     postproc->filter_pool_info = vi;
 
-    pool = gst_vaapi_surface_pool_new(postproc->display,
+    pool = gst_vaapi_surface_pool_new(GST_VAAPI_PLUGIN_BASE_DISPLAY(postproc),
         &postproc->filter_pool_info);
     if (!pool)
         return FALSE;
@@ -1440,14 +1199,13 @@ gst_vaapipostproc_set_caps(GstBaseTransform *trans, GstCaps *caps,
 
     if (caps_changed) {
         gst_vaapipostproc_destroy(postproc);
-        gst_caps_replace(&postproc->sinkpad_caps, caps);
-        gst_caps_replace(&postproc->srcpad_caps, out_caps);
         if (!gst_vaapipostproc_create(postproc))
             return FALSE;
+        if (!gst_vaapi_plugin_base_set_caps(GST_VAAPI_PLUGIN_BASE(trans),
+                caps, out_caps))
+            return FALSE;
     }
 
-    if (!ensure_sinkpad_buffer_pool(postproc, caps))
-        return FALSE;
     if (!ensure_srcpad_buffer_pool(postproc, out_caps))
         return FALSE;
     return TRUE;
@@ -1461,8 +1219,8 @@ gst_vaapipostproc_query(GstBaseTransform *trans, GstPadDirection direction,
 
     GST_INFO_OBJECT(trans, "query type `%s'", GST_QUERY_TYPE_NAME(query));
 
-    if (gst_vaapi_reply_to_query(query, postproc->display)) {
-        GST_DEBUG("sharing display %p", postproc->display);
+    if (gst_vaapi_reply_to_query(query, GST_VAAPI_PLUGIN_BASE_DISPLAY(postproc))) {
+        GST_DEBUG("sharing display %p", GST_VAAPI_PLUGIN_BASE_DISPLAY(postproc));
         return TRUE;
     }
 
@@ -1476,40 +1234,14 @@ gst_vaapipostproc_propose_allocation(GstBaseTransform *trans,
     GstQuery *decide_query, GstQuery *query)
 {
     GstVaapiPostproc * const postproc = GST_VAAPIPOSTPROC(trans);
-    GstCaps *caps = NULL;
-    gboolean need_pool;
+    GstVaapiPluginBase * const plugin = GST_VAAPI_PLUGIN_BASE(trans);
 
     /* Let vaapidecode allocate the video buffers */
     if (!postproc->is_raw_yuv)
         return FALSE;
-
-    gst_query_parse_allocation(query, &caps, &need_pool);
-
-    if (need_pool) {
-        if (!caps)
-            goto error_no_caps;
-        if (!ensure_sinkpad_buffer_pool(postproc, caps))
-            return FALSE;
-        gst_query_add_allocation_pool(query, postproc->sinkpad_buffer_pool,
-            postproc->sinkpad_buffer_size, 0, 0);
-    }
-
-    gst_query_add_allocation_meta(query,
-        GST_VAAPI_VIDEO_META_API_TYPE, NULL);
-    gst_query_add_allocation_meta(query,
-        GST_VIDEO_META_API_TYPE, NULL);
-    gst_query_add_allocation_meta(query,
-        GST_VIDEO_CROP_META_API_TYPE, NULL);
-    gst_query_add_allocation_meta(query,
-        GST_VIDEO_OVERLAY_COMPOSITION_META_API_TYPE, NULL);
-    return TRUE;
-
-    /* ERRORS */
-error_no_caps:
-    {
-        GST_ERROR("no caps specified");
+    if (!gst_vaapi_plugin_base_propose_allocation(plugin, query))
         return FALSE;
-    }
+    return TRUE;
 }
 #endif
 
@@ -1520,6 +1252,7 @@ gst_vaapipostproc_finalize(GObject *object)
 
     gst_vaapipostproc_destroy(postproc);
 
+    gst_vaapi_plugin_base_finalize(GST_VAAPI_PLUGIN_BASE(postproc));
     G_OBJECT_CLASS(gst_vaapipostproc_parent_class)->finalize(object);
 }
 
@@ -1560,6 +1293,22 @@ gst_vaapipostproc_set_property(
          postproc->sharpen_level = g_value_get_float(value);
          postproc->flags |= GST_VAAPI_POSTPROC_FLAG_SHARPEN;
          break;
+    case PROP_HUE:
+        postproc->hue = g_value_get_float(value);
+        postproc->flags |= GST_VAAPI_POSTPROC_FLAG_HUE;
+        break;
+    case PROP_SATURATION:
+        postproc->saturation = g_value_get_float(value);
+        postproc->flags |= GST_VAAPI_POSTPROC_FLAG_SATURATION;
+        break;
+    case PROP_BRIGHTNESS:
+        postproc->brightness = g_value_get_float(value);
+        postproc->flags |= GST_VAAPI_POSTPROC_FLAG_BRIGHTNESS;
+        break;
+    case PROP_CONTRAST:
+        postproc->contrast = g_value_get_float(value);
+        postproc->flags |= GST_VAAPI_POSTPROC_FLAG_CONTRAST;
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
@@ -1601,6 +1350,18 @@ gst_vaapipostproc_get_property(
     case PROP_SHARPEN:
         g_value_set_float(value, postproc->sharpen_level);
         break;
+    case PROP_HUE:
+        g_value_set_float(value, postproc->hue);
+        break;
+    case PROP_SATURATION:
+        g_value_set_float(value, postproc->saturation);
+        break;
+    case PROP_BRIGHTNESS:
+        g_value_set_float(value, postproc->brightness);
+        break;
+    case PROP_CONTRAST:
+        g_value_set_float(value, postproc->contrast);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
@@ -1620,6 +1381,8 @@ gst_vaapipostproc_class_init(GstVaapiPostprocClass *klass)
     GST_DEBUG_CATEGORY_INIT(gst_debug_vaapipostproc,
                             GST_PLUGIN_NAME, 0, GST_PLUGIN_DESC);
 
+    gst_vaapi_plugin_base_class_init(GST_VAAPI_PLUGIN_BASE_CLASS(klass));
+
     object_class->finalize      = gst_vaapipostproc_finalize;
     object_class->set_property  = gst_vaapipostproc_set_property;
     object_class->get_property  = gst_vaapipostproc_get_property;
@@ -1638,10 +1401,6 @@ gst_vaapipostproc_class_init(GstVaapiPostprocClass *klass)
     trans_class->prepare_output_buffer =
         gst_vaapipostproc_prepare_output_buffer;
 
-#if GST_CHECK_VERSION(1,1,0)
-    element_class->set_context = gst_vaapipostproc_set_context;
-#endif
-
     gst_element_class_set_static_metadata(element_class,
         "VA-API video postprocessing",
         "Filter/Converter/Video",
@@ -1716,7 +1475,7 @@ gst_vaapipostproc_class_init(GstVaapiPostprocClass *klass)
                            "Width",
                            "Forced output width",
                            0, G_MAXINT, 0,
-                           G_PARAM_READWRITE));
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
     /**
      * GstVaapiPostproc:height:
@@ -1732,7 +1491,7 @@ gst_vaapipostproc_class_init(GstVaapiPostprocClass *klass)
                            "Height",
                            "Forced output height",
                            0, G_MAXINT, 0,
-                           G_PARAM_READWRITE));
+                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
     /**
      * GstVaapiPostproc:force-aspect-ratio:
@@ -1748,7 +1507,7 @@ gst_vaapipostproc_class_init(GstVaapiPostprocClass *klass)
                               "Force aspect ratio",
                               "When enabled, scaling will respect original aspect ratio",
                               TRUE,
-                              G_PARAM_READWRITE));
+                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
     /**
      * GstVaapiPostproc:denoise:
@@ -1771,12 +1530,58 @@ gst_vaapipostproc_class_init(GstVaapiPostprocClass *klass)
         g_object_class_install_property(object_class,
             PROP_SHARPEN, filter_op->pspec);
 
+    /**
+     * GstVaapiPostproc:hue:
+     *
+     * The color hue, expressed as a float value. Range is -180.0 to
+     * 180.0. Default value is 0.0 and represents no modification.
+     */
+    filter_op = find_filter_op(filter_ops, GST_VAAPI_FILTER_OP_HUE);
+    if (filter_op)
+        g_object_class_install_property(object_class,
+            PROP_HUE, filter_op->pspec);
+
+    /**
+     * GstVaapiPostproc:saturation:
+     *
+     * The color saturation, expressed as a float value. Range is 0.0
+     * to 2.0. Default value is 1.0 and represents no modification.
+     */
+    filter_op = find_filter_op(filter_ops, GST_VAAPI_FILTER_OP_SATURATION);
+    if (filter_op)
+        g_object_class_install_property(object_class,
+            PROP_SATURATION, filter_op->pspec);
+
+    /**
+     * GstVaapiPostproc:brightness:
+     *
+     * The color brightness, expressed as a float value. Range is -1.0
+     * to 1.0. Default value is 0.0 and represents no modification.
+     */
+    filter_op = find_filter_op(filter_ops, GST_VAAPI_FILTER_OP_BRIGHTNESS);
+    if (filter_op)
+        g_object_class_install_property(object_class,
+            PROP_BRIGHTNESS, filter_op->pspec);
+
+    /**
+     * GstVaapiPostproc:contrast:
+     *
+     * The color contrast, expressed as a float value. Range is 0.0 to
+     * 2.0. Default value is 1.0 and represents no modification.
+     */
+    filter_op = find_filter_op(filter_ops, GST_VAAPI_FILTER_OP_CONTRAST);
+    if (filter_op)
+        g_object_class_install_property(object_class,
+            PROP_CONTRAST, filter_op->pspec);
+
     g_ptr_array_unref(filter_ops);
 }
 
 static void
 gst_vaapipostproc_init(GstVaapiPostproc *postproc)
 {
+    gst_vaapi_plugin_base_init(GST_VAAPI_PLUGIN_BASE(postproc), GST_CAT_DEFAULT);
+
     postproc->format                    = DEFAULT_FORMAT;
     postproc->deinterlace_mode          = DEFAULT_DEINTERLACE_MODE;
     postproc->deinterlace_method        = DEFAULT_DEINTERLACE_METHOD;
