@@ -43,8 +43,6 @@ struct _GstVaapiVideoMeta {
     GstVaapiDisplay            *display;
     GstVaapiVideoPool          *image_pool;
     GstVaapiImage              *image;
-    GstVaapiVideoPool          *surface_pool;
-    GstVaapiSurface            *surface;
     GstVaapiSurfaceProxy       *proxy;
     GFunc                       converter;
     guint                       render_flags;
@@ -73,31 +71,41 @@ set_image_from_pool(GstVaapiVideoMeta *meta, GstVaapiVideoPool *pool)
     image = gst_vaapi_video_pool_get_object(pool);
     if (!image)
         return FALSE;
+
     set_image(meta, image);
     meta->image_pool = gst_vaapi_video_pool_ref(pool);
     return TRUE;
 }
 
-static inline void
-set_surface(GstVaapiVideoMeta *meta, GstVaapiSurface *surface)
-{
-    meta->surface = gst_vaapi_object_ref(surface);
-    set_display(meta, gst_vaapi_object_get_display(GST_VAAPI_OBJECT(surface)));
-}
-
 static gboolean
-set_surface_from_pool(GstVaapiVideoMeta *meta, GstVaapiVideoPool *pool)
+set_surface_proxy(GstVaapiVideoMeta *meta, GstVaapiSurfaceProxy *proxy)
 {
     GstVaapiSurface *surface;
 
-    surface = gst_vaapi_video_pool_get_object(pool);
+    surface = GST_VAAPI_SURFACE_PROXY_SURFACE(proxy);
     if (!surface)
         return FALSE;
-    set_surface(meta, surface);
-    meta->surface_pool = gst_vaapi_video_pool_ref(pool);
+
+    meta->proxy = gst_vaapi_surface_proxy_ref(proxy);
+    set_display(meta, gst_vaapi_object_get_display(GST_VAAPI_OBJECT(surface)));
     return TRUE;
 }
 
+static gboolean
+set_surface_proxy_from_pool(GstVaapiVideoMeta *meta, GstVaapiVideoPool *pool)
+{
+    GstVaapiSurfaceProxy *proxy;
+    gboolean success;
+
+    proxy = gst_vaapi_surface_proxy_new_from_pool(GST_VAAPI_SURFACE_POOL(pool));
+    if (!proxy)
+        return FALSE;
+
+    success = set_surface_proxy(meta, proxy);
+    gst_vaapi_surface_proxy_unref(proxy);
+    return success;
+}
+
 static void
 gst_vaapi_video_meta_destroy_image(GstVaapiVideoMeta *meta)
 {
@@ -110,18 +118,10 @@ gst_vaapi_video_meta_destroy_image(GstVaapiVideoMeta *meta)
     gst_vaapi_video_pool_replace(&meta->image_pool, NULL);
 }
 
-static void
-gst_vaapi_video_meta_destroy_surface(GstVaapiVideoMeta *meta)
+static inline void
+gst_vaapi_video_meta_destroy_proxy(GstVaapiVideoMeta *meta)
 {
     gst_vaapi_surface_proxy_replace(&meta->proxy, NULL);
-
-    if (meta->surface) {
-        if (meta->surface_pool)
-            gst_vaapi_video_pool_put_object(meta->surface_pool, meta->surface);
-        gst_vaapi_object_unref(meta->surface);
-        meta->surface = NULL;
-    }
-    gst_vaapi_video_pool_replace(&meta->surface_pool, NULL);
 }
 
 #if !GST_CHECK_VERSION(1,0,0)
@@ -146,7 +146,7 @@ static void
 gst_vaapi_video_meta_finalize(GstVaapiVideoMeta *meta)
 {
     gst_vaapi_video_meta_destroy_image(meta);
-    gst_vaapi_video_meta_destroy_surface(meta);
+    gst_vaapi_video_meta_destroy_proxy(meta);
     gst_vaapi_display_replace(&meta->display, NULL);
 }
 
@@ -157,8 +157,6 @@ gst_vaapi_video_meta_init(GstVaapiVideoMeta *meta)
     meta->display       = NULL;
     meta->image_pool    = NULL;
     meta->image         = NULL;
-    meta->surface_pool  = NULL;
-    meta->surface       = NULL;
     meta->proxy         = NULL;
     meta->converter     = NULL;
     meta->render_flags  = 0;
@@ -217,7 +215,7 @@ gst_vaapi_video_meta_copy(GstVaapiVideoMeta *meta)
 
     g_return_val_if_fail(GST_VAAPI_IS_VIDEO_META(meta), NULL);
 
-    if (meta->image_pool || meta->surface_pool)
+    if (meta->image_pool)
         return NULL;
 
     copy = _gst_vaapi_video_meta_create();
@@ -228,10 +226,8 @@ gst_vaapi_video_meta_copy(GstVaapiVideoMeta *meta)
     copy->display       = gst_vaapi_display_ref(meta->display);
     copy->image_pool    = NULL;
     copy->image         = meta->image ? gst_vaapi_object_ref(meta->image) : NULL;
-    copy->surface_pool  = NULL;
-    copy->surface       = meta->surface ? gst_vaapi_object_ref(meta->surface) : NULL;
     copy->proxy         = meta->proxy ?
-        gst_vaapi_surface_proxy_ref(meta->proxy) : NULL;
+        gst_vaapi_surface_proxy_copy(meta->proxy) : NULL;
     copy->converter     = meta->converter;
     copy->render_flags  = meta->render_flags;
 
@@ -286,6 +282,7 @@ GstVaapiVideoMeta *
 gst_vaapi_video_meta_new_from_pool(GstVaapiVideoPool *pool)
 {
     GstVaapiVideoMeta *meta;
+    GstVaapiVideoPoolObjectType object_type;
 
     g_return_val_if_fail(pool != NULL, NULL);
 
@@ -293,15 +290,19 @@ gst_vaapi_video_meta_new_from_pool(GstVaapiVideoPool *pool)
     if (G_UNLIKELY(!meta))
         return NULL;
 
-    switch (gst_vaapi_video_pool_get_object_type(pool)) {
+    object_type = gst_vaapi_video_pool_get_object_type(pool);
+    switch (object_type) {
     case GST_VAAPI_VIDEO_POOL_OBJECT_TYPE_IMAGE:
         if (!set_image_from_pool(meta, pool))
             goto error;
         break;
     case GST_VAAPI_VIDEO_POOL_OBJECT_TYPE_SURFACE:
-        if (!set_surface_from_pool(meta, pool))
+        if (!set_surface_proxy_from_pool(meta, pool))
             goto error;
         break;
+    default:
+        GST_ERROR("unsupported video buffer pool of type %d", object_type);
+        goto error;
     }
     set_display(meta, gst_vaapi_video_pool_get_display(pool));
     return meta;
@@ -338,33 +339,6 @@ gst_vaapi_video_meta_new_with_image(GstVaapiImage *image)
     return meta;
 }
 
-/**
- * gst_vaapi_video_meta_new_with_surface:
- * @surface: a #GstVaapiSurface
- *
- * Creates a #GstVaapiVideoMeta with the specified @surface. The resulting
- * meta holds an additional reference to the @surface.
- *
- * This function shall only be called from within gstreamer-vaapi
- * plugin elements.
- *
- * Return value: the newly allocated #GstVaapiVideoMeta, or %NULL on error
- */
-GstVaapiVideoMeta *
-gst_vaapi_video_meta_new_with_surface(GstVaapiSurface *surface)
-{
-    GstVaapiVideoMeta *meta;
-
-    g_return_val_if_fail(surface != NULL, NULL);
-
-    meta = _gst_vaapi_video_meta_new();
-    if (G_UNLIKELY(!meta))
-        return NULL;
-
-    gst_vaapi_video_meta_set_surface(meta, surface);
-    return meta;
-}
-
 /**
  * gst_vaapi_video_meta_new_with_surface_proxy:
  * @proxy: a #GstVaapiSurfaceProxy
@@ -557,53 +531,7 @@ gst_vaapi_video_meta_get_surface(GstVaapiVideoMeta *meta)
 {
     g_return_val_if_fail(GST_VAAPI_IS_VIDEO_META(meta), NULL);
 
-    return meta->surface;
-}
-
-/**
- * gst_vaapi_video_meta_set_surface:
- * @meta: a #GstVaapiVideoMeta
- * @surface: a #GstVaapiSurface
- *
- * Binds @surface to the @meta. If the @meta contains another
- * surface previously allocated from a pool, it's pushed back to its
- * parent pool and the pool is also released.
- */
-void
-gst_vaapi_video_meta_set_surface(GstVaapiVideoMeta *meta,
-    GstVaapiSurface *surface)
-{
-    g_return_if_fail(GST_VAAPI_IS_VIDEO_META(meta));
-
-    gst_vaapi_video_meta_destroy_surface(meta);
-
-    if (surface)
-        set_surface(meta, surface);
-}
-
-/**
- * gst_vaapi_video_meta_set_surface_from_pool
- * @meta: a #GstVaapiVideoMeta
- * @pool: a #GstVaapiVideoPool
- *
- * Binds a newly allocated video object from the @pool. The @pool
- * shall be of type #GstVaapiSurfacePool. Previously allocated objects
- * are released and returned to their parent pools, if any.
- *
- * Return value: %TRUE on success
- */
-gboolean
-gst_vaapi_video_meta_set_surface_from_pool(GstVaapiVideoMeta *meta,
-    GstVaapiVideoPool *pool)
-{
-    g_return_val_if_fail(GST_VAAPI_IS_VIDEO_META(meta), FALSE);
-    g_return_val_if_fail(pool != NULL, FALSE);
-    g_return_val_if_fail(gst_vaapi_video_pool_get_object_type(pool) ==
-        GST_VAAPI_VIDEO_POOL_OBJECT_TYPE_SURFACE, FALSE);
-
-    gst_vaapi_video_meta_destroy_surface(meta);
-
-    return set_surface_from_pool(meta, pool);
+    return meta->proxy ? GST_VAAPI_SURFACE_PROXY_SURFACE(meta->proxy) : NULL;
 }
 
 /**
@@ -638,19 +566,15 @@ void
 gst_vaapi_video_meta_set_surface_proxy(GstVaapiVideoMeta *meta,
     GstVaapiSurfaceProxy *proxy)
 {
-    GstVaapiSurface *surface;
     const GstVaapiRectangle *crop_rect;
 
     g_return_if_fail(GST_VAAPI_IS_VIDEO_META(meta));
 
-    gst_vaapi_video_meta_destroy_surface(meta);
+    gst_vaapi_video_meta_destroy_proxy(meta);
 
     if (proxy) {
-        surface = GST_VAAPI_SURFACE_PROXY_SURFACE(proxy);
-        if (!surface)
+        if (!set_surface_proxy(meta, proxy))
             return;
-        set_surface(meta, surface);
-        meta->proxy = gst_vaapi_surface_proxy_ref(proxy);
 
         crop_rect = gst_vaapi_surface_proxy_get_crop_rect(proxy);
         if (crop_rect)
@@ -701,7 +625,7 @@ guint
 gst_vaapi_video_meta_get_render_flags(GstVaapiVideoMeta *meta)
 {
     g_return_val_if_fail(GST_VAAPI_IS_VIDEO_META(meta), 0);
-    g_return_val_if_fail(meta->surface != NULL, 0);
+    g_return_val_if_fail(meta->proxy != NULL, 0);
 
     return meta->render_flags;
 }
@@ -717,7 +641,7 @@ void
 gst_vaapi_video_meta_set_render_flags(GstVaapiVideoMeta *meta, guint flags)
 {
     g_return_if_fail(GST_VAAPI_IS_VIDEO_META(meta));
-    g_return_if_fail(meta->surface != NULL);
+    g_return_if_fail(meta->proxy != NULL);
 
     meta->render_flags = flags;
 }
