@@ -31,10 +31,16 @@
 
 G_BEGIN_DECLS
 
+typedef struct _GstVaapiVideoMetaTexture        GstVaapiVideoMetaTexture;
+
 G_GNUC_INTERNAL
 gboolean
 gst_buffer_add_texture_upload_meta(GstBuffer *buffer);
 
+G_GNUC_INTERNAL
+gboolean
+gst_buffer_ensure_texture_upload_meta(GstBuffer *buffer);
+
 G_END_DECLS
 
 #endif /* GST_VAAPI_VIDEO_META_TEXTURE_H */
